
/**************************************************************************
 *                                                                        *
 * This code is developed by Adam Li.  This software is an                *
 * implementation of a part of one or more MPEG-4 Video tools as          *
 * specified in ISO/IEC 14496-2 standard.  Those intending to use this    *
 * software module in hardware or software products are advised that its  *
 * use may infringe existing patents or copyrights, and any such use      *
 * would be at such party's own risk.  The original developer of this     *
 * software module and his/her company, and subsequent editors and their  *
 * companies (including Project Mayo), will have no liability for use of  *
 * this software or modifications or derivatives thereof.                 *
 *                                                                        *
 * Project Mayo gives users of the Codec a license to this software       *
 * module or modifications thereof for use in hardware or software        *
 * products claiming conformance to the MPEG-4 Video Standard as          *
 * described in the Open DivX license.                                    *
 *                                                                        *
 * The complete Open DivX license can be found at                         *
 * http://www.projectmayo.com/opendivx/license.php .                      *
 *                                                                        *
 **************************************************************************/

/************************************************************************
 *
 *  dec_engine.cpp, Decoder Engines
 *
 *  Copyright (C) 2000  DivX Networks
 * 
 *  Adam Li
 *  Andrea Graziani
 *  Cyril
 *
 *  DivX Advance Research Center <darc@projectmayo.com>
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "stdafx.h"
#include "codec.h"
#include "decore.h"


FOURCC_LOOKUP dec_lookup_table[] = 
{
    mmioFOURCC('Y','V','1','2'), DEC_420, 
		mmioFOURCC('Y','U','Y','2'), DEC_YUV2 
    , 0,0    // end of table marker
};


long codec::decBegin(LPARAM lParam1, LPARAM lParam2)
{
	BITMAPINFO *lpbiInput, *lpbiOutput;
	BITMAPV4HEADER *infohdr_i, *infohdr_o;
	DEC_MEM_REQS decMemReqs;
	DEC_PARAM decParam;
	DEC_SET decSet;

	unsigned int output_format;
	char * pnColor;
	unsigned long * pdwColor;
	FOURCC_LOOKUP * pFOURCC_LOOKUP;

	lpbiInput = (BITMAPINFO *)lParam1;
	lpbiOutput = (BITMAPINFO *)lParam2;
	infohdr_i = (BITMAPV4HEADER *)&(lpbiInput->bmiHeader);
	infohdr_o = (BITMAPV4HEADER *)&(lpbiOutput->bmiHeader);

	decParam.x_dim = infohdr_i->bV4Width;
	decParam.y_dim = abs(infohdr_i->bV4Height);
	decParam.output_format = 0;

	output_format = 0;
	
	if (infohdr_o->bV4V4Compression == BI_RGB) 
	{
		if ( infohdr_o->bV4BitCount == 32 ) {
			if ( infohdr_o->bV4Height >= 0 ) {
				output_format = DEC_RGB32;
			} else {
				output_format = DEC_RGB32_INV;
			}
		}
		if ( infohdr_o->bV4BitCount == 24 ) {
			if ( infohdr_o->bV4Height >= 0 ) {
				output_format = DEC_RGB24;
			} else {
				output_format = DEC_RGB24_INV;
			}
		}
		if ( infohdr_o->bV4BitCount == 16 ) {
			if ( infohdr_o->bV4Height >= 0 ) {
				output_format = DEC_RGB555;
			} else {
				output_format = DEC_RGB555_INV;
			}
		}
	} 
	else if (infohdr_o->bV4V4Compression == BI_BITFIELDS) 
	{
		if ( infohdr_o->bV4BitCount == 16 ) {
			// this is the way MS docs says pointer to RGBQUAD should be computed
	    pnColor = ((LPSTR)lpbiOutput + (WORD)(lpbiOutput->bmiHeader.biSize)); 
			pdwColor = (unsigned long*) pnColor;
			if ( (pdwColor[0] == 0x00f800) && (pdwColor[1] == 0x07e0) && (pdwColor[2] == 0x001f) ) {
				if ( infohdr_o->bV4Height >= 0 ) {
					output_format = DEC_RGB565;
				} else {
					output_format = DEC_RGB565_INV;
				}
			}
		}
	} 
	else 
	{
		pFOURCC_LOOKUP = dec_lookup_table;
		while ( pFOURCC_LOOKUP->nFourCC ) {
			if ( infohdr_o->bV4V4Compression == pFOURCC_LOOKUP->nFourCC ) {
				output_format = pFOURCC_LOOKUP->nInternalCode;
				break;
			}
			pFOURCC_LOOKUP++;
		}
	}
	if ( output_format == 0 ) {
		return ICERR_BADFORMAT;
	}
	
	// ok got a valid output format so save it to param block
	decParam.output_format = output_format;

  // since VFW specification allows an application to call 'Begin'
  // and then call 'Begin' again without first calling 'End'
  if ( m_bDecompressBeginDone ) {
    decore((long) this, DEC_OPT_RELEASE, NULL, NULL);
      m_bDecompressBeginDone = false;
  }

	decore((long) this, DEC_OPT_MEMORY_REQS, &decParam, &decMemReqs);

	decParam.buffers.mp4_edged_ref_buffers = malloc(decMemReqs.mp4_edged_ref_buffers_size);
	decParam.buffers.mp4_edged_for_buffers = malloc(decMemReqs.mp4_edged_for_buffers_size);
	decParam.buffers.mp4_display_buffers = malloc(decMemReqs.mp4_display_buffers_size);
	decParam.buffers.mp4_state = malloc(decMemReqs.mp4_state_size);
	decParam.buffers.mp4_tables = malloc(decMemReqs.mp4_tables_size);
	decParam.buffers.mp4_stream = malloc(decMemReqs.mp4_stream_size);

	memset(decParam.buffers.mp4_state, 0, decMemReqs.mp4_state_size);
	memset(decParam.buffers.mp4_tables, 0, decMemReqs.mp4_tables_size);
	memset(decParam.buffers.mp4_stream, 0, decMemReqs.mp4_stream_size);


	decore((long) this, DEC_OPT_INIT, &decParam, NULL);

	// The postprocessing of the decore can be hard coded here for now.
	// The filter that produces message will be implemented later.

	// postprcc_level =  0      ---->  no post processing
	// postproc_level =  1 ~  9 ---->  horizontal deblocking only
	// postproc_level = 10 ~ 19 ---->  hor. + ver. deblocking
	// postproc_level = 20 ~ 29 ---->  full deblocking + deringing
	// postproc_level = 30 ~ 39 ---->  above + deblocking chroma hor.
	// postproc_level = 40 ~ 49 ---->  above + deblocking chroma ver.
	// postproc_level = 50 ~ 59 ---->  above + deringing chroma

	decSet.postproc_level = 0;

	decore((long) this, DEC_OPT_SETPP, &decSet, NULL);

  m_bDecompressBeginDone = true;

	return ICERR_OK;
}

long codec::decEnd(LPARAM lParam1, LPARAM lParam2)
{
  // since VirtualDub v1.4c calls this without doing a Begin
  if ( m_bDecompressBeginDone ) {
    decore((long) this, DEC_OPT_RELEASE, NULL, NULL);
      m_bDecompressBeginDone = false;
  }

	decore((long) this, DEC_OPT_RELEASE, NULL, NULL);

	return ICERR_OK;
}

long codec::decDecode(LPARAM lParam1, LPARAM lParam2)
{
	ICDECOMPRESS *icc = (ICDECOMPRESS *)lParam1;
	ICOPEN *ico = (ICOPEN *)lParam2;

	DWORD flags = icc->dwFlags;

	BITMAPV4HEADER *lpbihdr_i, *lpbihdr_o;
	DEC_FRAME dec_frame;

	lpbihdr_i = (BITMAPV4HEADER *)icc->lpbiInput;
	lpbihdr_o = (BITMAPV4HEADER *)icc->lpbiOutput;

	dec_frame.length = icc->lpbiInput->biSizeImage;
	dec_frame.bitstream = icc->lpInput;
	dec_frame.bmp = icc->lpOutput;
	dec_frame.stride = lpbihdr_o->bV4Width;

	if ((flags & ICDECOMPRESS_HURRYUP) | (flags & ICDECOMPRESS_UPDATE)) {
		dec_frame.render_flag = 0;
	}
	else {
		dec_frame.render_flag = 1;
	}

	decore((long) this, 0, &dec_frame, NULL);
 
	return ICERR_OK;
}

long codec::decSetPostProcessing(LPARAM lParam1, LPARAM lParam2)
{
	DEC_SET *dec_set = (DEC_SET *)lParam1;

	decore((long) this, DEC_OPT_SETPP, dec_set, NULL);

	return ICERR_OK;
}
